from cnvrg.modules.data_connector.base_connector import BaseConnector

class BaseFileConnector(BaseConnector):
    def __init__(self, data_connector):
        super(BaseFileConnector, self).__init__(data_connector)
        self._files_callback = lambda x: x

    def __len__(self):
        raise Exception("Not implemented")

    def __getitem__(self, item):
        raise Exception("Not implemented")

    @property
    def working_dir(self):
        raise Exception("Not implemented")


    def torch_ds(self):
        from torch.utils.data.dataset import Dataset as TorchDataset
        class TorchDs(TorchDataset):
            def __init__(self, dc):
                self.__dc = dc
            def __getitem__(self, item):
                return self.__dc[item]
            def __len__(self):
                return len(self.__dc)
        return TorchDs(self)

    def data_loader(self, **kwargs):
        # Arguments:
        #         dataset (Dataset): dataset from which to load the data.
        #         batch_size (int, optional): how many samples per batch to load
        #             (default: ``1``).
        #         shuffle (bool, optional): set to ``True`` to have the data reshuffled
        #             at every epoch (default: ``False``).
        #         sampler (Sampler, optional): defines the strategy to draw samples from
        #             the dataset. If specified, ``shuffle`` must be False.
        #         batch_sampler (Sampler, optional): like sampler, but returns a batch of
        #             indices at a time. Mutually exclusive with :attr:`batch_size`,
        #             :attr:`shuffle`, :attr:`sampler`, and :attr:`drop_last`.
        #         num_workers (int, optional): how many subprocesses to use for data
        #             loading. 0 means that the data will be loaded in the main process.
        #             (default: ``0``)
        #         collate_fn (callable, optional): merges a list of samples to form a mini-batch.
        #         pin_memory (bool, optional): If ``True``, the data loader will copy tensors
        #             into CUDA pinned memory before returning them.  If your data elements
        #             are a custom type, or your ``collate_fn`` returns a batch that is a custom type
        #             see the example below.
        #         drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
        #             if the dataset size is not divisible by the batch size. If ``False`` and
        #             the size of dataset is not divisible by the batch size, then the last batch
        #             will be smaller. (default: ``False``)
        #         timeout (numeric, optional): if positive, the timeout value for collecting a batch
        #             from workers. Should always be non-negative. (default: ``0``)
        #         worker_init_fn (callable, optional): If not ``None``, this will be called on each
        #             worker subprocess with the worker id (an int in ``[0, num_workers - 1]``) as
        #             input, after seeding and before data loading. (default: ``None``)
        #         sort_fn (list) -> list: a callback which get a list of files to download and should return a sorted list of
        #           files, (default lambda x: x)
        from torch.utils import data
        class DataLoader(data.DataLoader):
            def __init__(self, dc, **kwargs):
                super(DataLoader, self).__init__(dc.torch_ds(), **kwargs)
        self._files_callback = kwargs.get("sort_fn") or (lambda x: x)
        if 'sort_fn' in kwargs: del kwargs["sort_fn"]
        return DataLoader(self, **kwargs)


from .base_sql_connector import BaseSqlConnector
from cnvrg.modules.errors import CnvrgError
is_loaded = False
try:
    import thrift_sasl
    from pyhive import hive
    is_loaded = True

except ImportError:
    is_loaded = False


class HiveConnector(BaseSqlConnector):

    @staticmethod
    def requirements():
        return """thrift_sasl\npyhive\nthrift"""


    @staticmethod
    def key_type():
        return "hive"

    def __init__(self, data_connector):
        if not is_loaded: raise CnvrgError("Cant find pyhive module, please install it before using HiveConnector")
        super(HiveConnector, self).__init__(data_connector)

    def _connection(self):
        # host=None, port=None, username=None, database='default', auth=None,
        #                  configuration=None, kerberos_service_name=None, password=None,
        #                  thrift_transport=None
        return hive.Connection(**self.data)



from cnvrg.modules.cnvrg_job import CnvrgJob, LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR
import cnvrg.helpers.param_build_helper as param_build_helper
from cnvrg.modules.project import Project
import cnvrg.helpers.logger_helper as logger_helper
import cnvrg.helpers.string_helper as string_helper
from cnvrg.modules.errors import UserError
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.spawn_helper as spawn_helper
import cnvrg.helpers.control_stdout_helper as control_stdout_helper
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.chart_show_helper as chart_show_helper
from cnvrg.helpers.env_helper import in_experiment, CURRENT_JOB_ID
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.param_helper import wrap_string_to_list
from cnvrg.helpers.libs_helper import check_if_lib_exists
from enum import Enum
from typing import List
import numpy as np
import types
import os
import time


class TagType(Enum):
    SINGLE_TAG = "single"
    LINECHART_TAG = "linechart"



class Experiment(CnvrgJob):
    def __init__(self, experiment=None):
        owner, project_slug, slug = param_build_helper.parse_params(experiment, param_build_helper.EXPERIMENT)
        if not in_experiment() and not slug:
            raise UserError("Cant create an experiment without slug and outside experiment run")
        slug = slug or CURRENT_JOB_ID
        super(Experiment, self).__init__(slug, env_helper.EXPERIMENT, Project(url_join(owner, project_slug)))
        self.title = None
        self.__data = self.__get_experiment()

    @staticmethod
    def init(project=None, **kwargs):
        project = project or Project()
        resp = apis_helper.post(url_join(project.get_base_url(), 'experiments', 'local'),
                                data={"commit": project.get_current_commit(), **kwargs, **env_helper.get_origin_job()})
        e = resp.get("experiment")
        exp = Experiment(url_join(project.get_project_name(), e.get("slug")))
        logger_helper.log_message("{title} is running, follow live at: {url}".format(title=exp['title'], url=exp.href()))
        return exp

    @staticmethod
    def __run_callable(function_command, title=None, project=None, cwd=None, library=None, **kwargs):
        experiment = Experiment.init(title=title, project=project, library=library, **kwargs)
        exit_status = control_stdout_helper.run_callable(function_command, arguments=[experiment], callback=lambda x: experiment.log(x), err_callback=lambda x: experiment.log(x, log_type=LOGS_TYPE_ERROR))
        experiment.finish(exit_status=exit_status)
        return experiment

    @staticmethod
    def __run_local(cmd, **kwargs):
        function_command = lambda *args: spawn_helper.run_sync(cmd, print_output=True, cwd=kwargs["cwd"])
        return Experiment.__run_callable(function_command, cmd=cmd, **kwargs)


    @staticmethod
    def sync(**kwargs):
        Project().sync(**{**CnvrgJob.current_job_sync_args(), **kwargs})

    @staticmethod
    def run(command=None, title=None, project=None, compute=None, computes=None, datasets=None, dataset=None, local=True, library=None, working_directory=None, sync_before=True, sync_after=True, **kwargs):
        function_command = None
        project = project or Project()

        ## sync before
        if sync_before: project.sync(**CnvrgJob.current_job_sync_args())

        ## support
        computes = computes or wrap_string_to_list(computes) or wrap_string_to_list(compute)
        ### by default (if computes empty or no compute) set the local to True
        if not computes: local = True
        elif "local" in computes: local = True
        elif len(computes) == 0: local = True

        ### if command is function, run it as a function and not as a os command
        if isinstance(command, types.FunctionType): function_command = command

        ## merge datasets
        datasets = datasets or wrap_string_to_list(datasets) or wrap_string_to_list(dataset)




        if function_command:
            ## set the title to "func: {name of the function}"
            cmd = "func: {func_name}".format(func_name=function_command.__name__)
            experiment = Experiment.__run_callable(function_command, title=title, project=project, library=library, cmd=cmd)
            if not experiment: raise UserError("Cant run experiment, please check your input params")
        elif local:

            ## if its a string command && local=True, run it as a local task (subprocess.Popen)
            experiment = Experiment.__run_local(cmd=command, title=title, project=project, cwd=working_directory, library=library)
            if not experiment: raise UserError("Cant run experiment, please check your input params")
        else:

            ## perform an api call to create it.
            e = project.run_task(command, title=title, templates=computes, datasets=datasets, library=library, **kwargs)
            return Experiment(e.get("experiments")[0])

        if sync_after: project.sync(**CnvrgJob.current_job_sync_args(job=experiment['slug']))
        return experiment



    def log_param(self, key, value=None):
        tag_data = {
            "key": key,
            "value": value,
            "type": TagType.SINGLE_TAG.value
        }
        self.__send_tag(tag_data)

    def __dict__(self):
        return self.__data

    def log_metric(self, key, Ys: List, Xs: List=None, grouping: List=None, x_axis=None, y_axis=None) -> None:
        """
        a function which can tag an experiment with a chart
        :param key: the name of the chart
        :param Ys: [y1, y2, y3, y4] (float)
        :param Xs: [x1, x2, x3, x4] (date, integer, null)
        :param grouping: [g1, g2, g3, g4]
        :param x_axis: rename the x_axis of the chart
        :param y_axis:rename the y_axis of the chart
        :return:
        """
        if isinstance(Xs, np.ndarray): Xs = Xs.tolist()
        if isinstance(Ys, np.ndarray): Ys = Ys.tolist()
        tag_data = {
            "ys": Ys,
            "xs": Xs,
            "key": key,
            "grouping": grouping,
            "x_axis": x_axis,
            "y_axis": y_axis,
            "type": TagType.LINECHART_TAG.value,
        }
        self.__send_tag(tag_data)

    def logs(self, callback=None, poll_every=5):
        job_logs, experiment_is_running = self.__fetch_logs(0)
        offset = len(job_logs)
        callback = callback or logger_helper.log_cnvrg_log
        [callback(l) for l in job_logs]
        while experiment_is_running:
            time.sleep(poll_every)
            job_logs, experiment_is_running = self.__fetch_logs(offset)
            offset += len(job_logs)
            ## filter unknown logs?
            [callback(l) if l else None for l in job_logs]


    @property
    def title(self):
        return self["title"]

    @title.setter
    def title(self, new_title):
        apis_helper.put(url_join(self._base_url(), 'title'), data={"title": new_title})
        self.__data = self.__get_experiment()

    def set_title(self,new_title):
        self.title = new_title


    def show_chart(self, key, **kwargs):
        """

        :param key: chart_key
        :param kwargs: with_legend, legend_loc
        :return:
        """
        chart = apis_helper.get(url_join(self._base_url(), 'charts', key)).get("chart")

        return chart_show_helper.show_chart(chart, **kwargs)


    def __fetch_logs(self, offset, limit=None):
        resp = apis_helper.get(url_join(self._base_url(), 'logs'), data={"offset": offset, "limit": limit})
        return resp.get("logs"), resp.get("experiment").get("is_running")


    def __send_tag(self,tag_data):
        apis_helper.post(url_join(self._base_url(), 'tags'), data={"tag": tag_data})

    def finish(self, exit_status=None):
        apis_helper.post(url_join(self._base_url(), 'finish'), data={"exit_status": exit_status})

    def href(self):
        return self.__data.get("full_href")

    def open(self):
        if check_if_lib_exists("webbrowser"):
            import webbrowser
            webbrowser.open(self.href())
        else:
            logger_helper.log_message("You can find the experiment in: {href}".format(href=self.href))


    def __get_experiment(self):
        return apis_helper.get(self._base_url()).get("experiment")

    def _base_url(self):
        return url_join(
            #### hackish :D
            self.project.get_base_url(),string_helper.to_snake_case(self.job_type) + "s", self.job_slug
        )

    def __getitem__(self, item):
        return self.__data.get(item)






from cnvrg.modules.data_connector.base_connector import BaseConnector as _BaseConnector
from cnvrg.modules.data_connector.files_connector import DatasetConnector, S3BucketConnector
from cnvrg.modules.data_connector.sql_connector import MysqlConnector, SnowflakeConnector, HiveConnector
from cnvrg.modules.data_connector.stream_connector import KafkaConnector



def DataConnector(data_connector):
    connector = _BaseConnector(data_connector)
    if connector.connector_type() == HiveConnector.key_type():
        return HiveConnector(data_connector)

    if connector.connector_type() == MysqlConnector.key_type():
        return MysqlConnector(data_connector)

    if connector.connector_type() == SnowflakeConnector.key_type():
        return SnowflakeConnector(data_connector)

    if connector.connector_type() == DatasetConnector.key_type():
        return DatasetConnector(data_connector)

    if connector.connector_type() == S3BucketConnector.key_type():
        return S3BucketConnector(data_connector)

    if connector.connector_type() == KafkaConnector.key_type():
        return KafkaConnector(data_connector)



from cnvrg.modules.data_connector.sql_connector.mysql_connector import MysqlConnector
from cnvrg.modules.data_connector.sql_connector.snowflake_connector import SnowflakeConnector
from cnvrg.modules.data_connector.sql_connector.hive_connector import HiveConnector


import click
import logging
import os
import json

LEVEL_INFO='blue'
LEVEL_ERROR="red"
LEVEL_SUCCESS="greed"

os.makedirs(os.path.join(os.path.expanduser("~"), '.cnvrg'), exist_ok=True)
logging.basicConfig(filename=os.path.join(os.path.expanduser("~"), '.cnvrg', 'cnvrg.log'), filemode='w', format='%(name)s - %(levelname)s - %(message)s')

def log_message(message, level=LEVEL_INFO, *args, **kwargs):
    try:
        click.secho(message, *args, **{"color": level, **kwargs})
    except Exception as e:
        print(message)
    logging.info(message)

def log_cnvrg_log(log):
    log_type = log.get("type")
    if log_type == "cnvrg-info":
        level = LEVEL_INFO
    elif log_type == "cnvrg-error":
        level = LEVEL_ERROR
    else:
        level = None
    log_message(log.get("message").strip(), level=level)


def log_bad_response(url, data=None, headers=None):
    logging.error("Got bad response to {url}, \nwith data: {data}, \nwith headers: {headers}".format(url=url, data=json.dumps(data or {}), headers=json.dumps(headers or {})))

def log_error(exception: Exception):
    logging.exception(exception)


from cnvrg.modules.base_module import CnvrgBase
import numpy as np
import itertools
from functools import reduce


class Grid(CnvrgBase):
    def __init__(self):
        pass

    def exec(self, object, **kwargs):
        pass

    @staticmethod
    def calculate_params(**kwargs):
        values = {k: Grid.__inflate_values(v) for k, v in kwargs.items()}
        values = [[{k: x} for x in v] for k, v in values.items()]
        return [reduce(lambda t, y: {**y, **t}, x) for x in itertools.product(*values)]

    @staticmethod
    def __inflate_values(definition):
        key_type = definition.get("type")
        if key_type == "array":
            return definition.get("values")

        if key_type == "range":
            d_start, d_stop = definition.get("start"), definition.get("stop")
            d_step = definition.get("step") or 1
            return np.arange(d_start, d_stop, d_step)

        if key_type == 'logspace':
            d_min, d_max, d_num = definition.get("start"), definition.get("stop"), definition.get("num")
            d_base = definition.get("base") or 10
            return np.logspace(d_min, d_max, num=d_num, base=d_base)


from .base_files_connector import BaseFileConnector
import cnvrg.modules.errors as errors
import os
import os

is_loaded = False
try:
    import boto3
    is_loaded = True

except ImportError:
    is_loaded = False



class S3BucketConnector(BaseFileConnector):

    @staticmethod
    def key_type():
        return "s3_bucket"

    def __init__(self, data_connector, prefix=None, working_dir=None):
        super(S3BucketConnector, self).__init__(data_connector)
        self.client = None
        self.__files = []
        self.__prefix = prefix or self.data.get("prefix")
        temp_working_dir = working_dir or "/data"
        if not os.access(temp_working_dir, os.W_OK): temp_working_dir = os.path.expanduser("~/cnvrg_datasets")
        self.__working_dir = os.path.join(temp_working_dir, self._data_connector)
        os.makedirs(self.__working_dir, exist_ok=True)
        self.__connect(**self.data)

    def __connect(self, access_key_id=None, secret_access_key=None, session_token=None, region=None, bucket=None, **kwargs):
        if not is_loaded: raise errors.CnvrgError("Cant load boto3 library.")
        self.data = {"aws_access_key_id": access_key_id, "aws_secret_access_key": secret_access_key, "aws_session_token": session_token, "region_name": region, "bucket": bucket}


    def get_bucket(self):
        return S3BucketConnector.bucket(**self.data)

    def get_client(self):
        return S3BucketConnector.client(**self.data)

    @staticmethod
    def client(**data):
        return boto3.client('s3', aws_access_key_id=data.get("key"), aws_secret_access_key=data.get("secret"), aws_session_token=data.get("session_token"), region_name=data.get("region"))

    @staticmethod
    def session(**data):
        return boto3.Session(aws_access_key_id=data.get("key"), aws_secret_access_key=data.get("secret"), aws_session_token=data.get("session_token"), region_name=data.get("region"))

    @staticmethod
    def bucket(bucket=None, **data):
        return S3BucketConnector.session(**data).resource('s3').Bucket(bucket)

    def list_files(self, prefix=None):
        self.__files = []
        prefix = prefix or self.__prefix
        for o in self.get_bucket().objects.filter(Prefix=prefix or ''):
            self.__files.append(o.key)

        ### we need the have a callback to let the user
        ### sort the files.
        return self._files_callback(self.__files)

    def __len__(self):
        if len(self.__files) == 0: self.list_files()
        return len(self.__files)

    def __getitem__(self, item):
        storage_path = self.__files[item]
        local_path = os.path.join(self.__working_dir, storage_path)
        return self.download_file(storage_path, local_path)

    @property
    def working_dir(self):
        return self.__working_dir

    def download_file(self, storage_path, local_path):
        client = S3BucketConnector.client(**self.data)
        if os.path.exists(local_path):
            return local_path
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        client.download_file(self.data.get("bucket"), storage_path, local_path)
        return local_path
